@@ -88,10 +88,9 @@ public class DefaultSftpClient extends AbstractSftpClient {
      *                                - if {@code null} the data is silently ignored
      * @throws IOException            If failed to initialize
      */
-    public DefaultSftpClient(
-                             ClientSession clientSession, SftpVersionSelector initialVersionSelector,
+    public DefaultSftpClient(ClientSession clientSession, SftpVersionSelector initialVersionSelector,
                              SftpErrorDataHandler errorDataHandler)
-                                                                    throws IOException {
+            throws IOException {
         super(errorDataHandler);
 
         this.nameDecodingCharset = SftpModuleProperties.NAME_DECODING_CHARSET.getRequired(clientSession);
@@ -282,14 +281,14 @@ public class DefaultSftpClient extends AbstractSftpClient {
             int s = buffer.rpos() - hdr;
             buffer.rpos(s);
             buffer.wpos(s);
-            buffer.putInt(1 /* cmd */ + Integer.BYTES /* id */ + len); // length
+            buffer.putUInt(1 /* cmd */ + Integer.BYTES /* id */ + len); // length
             buffer.putByte((byte) (cmd & 0xFF)); // cmd
-            buffer.putInt(id); // id
+            buffer.putInt(id);
             buffer.wpos(wpos);
             buf = buffer;
         } else {
             buf = new ByteArrayBuffer(hdr + len);
-            buf.putInt(1 /* cmd */ + Integer.BYTES /* id */ + len);
+            buf.putUInt(1 /* cmd */ + Integer.BYTES /* id */ + len);
             buf.putByte((byte) (cmd & 0xFF));
             buf.putInt(id);
             buf.putBuffer(buffer);
@@ -367,9 +366,9 @@ public class DefaultSftpClient extends AbstractSftpClient {
 
         // Send init packet
         Buffer buf = new ByteArrayBuffer(INIT_COMMAND_SIZE + SshConstants.SSH_PACKET_HEADER_LEN);
-        buf.putInt(INIT_COMMAND_SIZE);
+        buf.putUInt(INIT_COMMAND_SIZE);
         buf.putByte((byte) SftpConstants.SSH_FXP_INIT);
-        buf.putInt(initialVersion);
+        buf.putUInt(initialVersion);
 
         boolean traceEnabled = log.isTraceEnabled();
         ClientChannel clientChannel = getClientChannel();
@@ -389,50 +388,45 @@ public class DefaultSftpClient extends AbstractSftpClient {
 
     protected void handleInitResponse(Buffer buffer) throws IOException {
         boolean traceEnabled = log.isTraceEnabled();
+        SftpResponse response = SftpResponse.parse(SftpConstants.SSH_FXP_INIT, buffer);
         ClientChannel clientChannel = getClientChannel();
-        int length = buffer.getInt();
-        int type = buffer.getUByte();
-        int id = buffer.getInt();
+        int length = response.getLength();
+        int type = response.getType();
+        int id = response.getId();
         if (traceEnabled) {
             log.trace("handleInitResponse({}) id={} type={} len={}",
                     clientChannel, id, SftpConstants.getCommandMessageName(type), length);
         }
 
-        if (type == SftpConstants.SSH_FXP_VERSION) {
-            if ((id < SftpConstants.SFTP_V3) || (id > SftpConstants.SFTP_V6)) {
-                throw new SshException("Unsupported sftp version " + id);
-            }
-            versionHolder.set(id);
-
-            if (traceEnabled) {
-                log.trace("handleInitResponse({}) version={}", clientChannel, versionHolder);
-            }
-
-            while (buffer.available() > 0) {
-                String name = buffer.getString();
-                byte[] data = buffer.getBytes();
-                if (traceEnabled) {
-                    log.trace("handleInitResponse({}) added extension={}", clientChannel, name);
+        switch (type) {
+            case SftpConstants.SSH_FXP_VERSION:
+                if ((id < SftpConstants.SFTP_V3) || (id > SftpConstants.SFTP_V6)) {
+                    throw new SshException("Unsupported sftp version " + id);
                 }
-                extensions.put(name, data);
-            }
-        } else if (type == SftpConstants.SSH_FXP_STATUS) {
-            int substatus = buffer.getInt();
-            String msg = buffer.getString();
-            String lang = buffer.getString();
-            if (traceEnabled) {
-                log.trace("handleInitResponse({})[id={}] - status: {} [{}] {}",
-                        clientChannel, id, SftpConstants.getStatusName(substatus), lang, msg);
-            }
+                versionHolder.set(id);
 
-            throwStatusException(SftpConstants.SSH_FXP_INIT, id, substatus, msg, lang);
-        } else {
-            IOException err = handleUnexpectedPacket(
-                    SftpConstants.SSH_FXP_INIT, SftpConstants.SSH_FXP_VERSION, id, type, length, buffer);
-            if (err != null) {
-                throw err;
-            }
+                if (traceEnabled) {
+                    log.trace("handleInitResponse({}) version={}", clientChannel, versionHolder);
+                }
 
+                while (buffer.available() > 0) {
+                    String name = buffer.getString();
+                    byte[] data = buffer.getBytes();
+                    if (traceEnabled) {
+                        log.trace("handleInitResponse({}) added extension={}", clientChannel, name);
+                    }
+                    extensions.put(name, data);
+                }
+                break;
+            case SftpConstants.SSH_FXP_STATUS:
+                throwStatusException(SftpConstants.SSH_FXP_INIT, id, SftpStatus.parse(response));
+                break;
+            default:
+                IOException err = handleUnexpectedPacket(SftpConstants.SSH_FXP_VERSION, response);
+                if (err != null) {
+                    throw err;
+                }
+                break;
         }
     }
 
@@ -494,7 +488,7 @@ public class DefaultSftpClient extends AbstractSftpClient {
 
         Map<String, ?> parsed = getParsedServerExtensions();
         Collection<String> extensions = ParserUtils.supportedExtensions(parsed);
-        List<Integer> availableVersions = Collections.emptyList();
+        List<Integer> availableVersions;
         if ((GenericUtils.size(extensions) > 0)
                 && extensions.contains(SftpConstants.EXT_VERSION_SELECT)) {
             Versions vers = MapEntryUtils.isEmpty(parsed)
@@ -551,13 +545,14 @@ public class DefaultSftpClient extends AbstractSftpClient {
             boolean wantReply = CoreModuleProperties.REQUEST_SUBSYSTEM_REPLY.getRequired(this);
             Buffer buffer = session.createBuffer(SshConstants.SSH_MSG_CHANNEL_REQUEST,
                     Channel.CHANNEL_SUBSYSTEM.length() + systemName.length() + Integer.SIZE);
-            buffer.putInt(getRecipient());
+            buffer.putUInt(getRecipient());
             buffer.putString(Channel.CHANNEL_SUBSYSTEM);
             buffer.putBoolean(wantReply);
             buffer.putString(systemName);
             addPendingRequest(Channel.CHANNEL_SUBSYSTEM, wantReply);
             writePacket(buffer);
 
+            setStreaming(Streaming.Async);
             asyncIn = createAsyncInput(session);
             setOut(createStdOutputStream(session));
             setErr(createErrOutputStream(session));
@@ -565,17 +560,21 @@ public class DefaultSftpClient extends AbstractSftpClient {
 
         protected ChannelAsyncOutputStream createAsyncInput(Session session) {
             return new ChannelAsyncOutputStream(
-                    this, SshConstants.SSH_MSG_CHANNEL_DATA,
-                    SftpModuleProperties.CHUNK_IF_WINDOW_LESS_THAN_PACKET.getRequired(session)) {
+                    this, SshConstants.SSH_MSG_CHANNEL_DATA) {
                 @SuppressWarnings("synthetic-access")
                 @Override
                 protected CloseFuture doCloseGracefully() {
-                    try {
-                        sendEof();
-                    } catch (IOException e) {
-                        session.exceptionCaught(e);
-                    }
-                    return super.doCloseGracefully();
+                    // First get the last packets out
+                    CloseFuture result = super.doCloseGracefully();
+                    result.addListener(f -> {
+                        try {
+                            // The channel writes EOF directly through the SSH session
+                            sendEof();
+                        } catch (IOException e) {
+                            session.exceptionCaught(e);
+                        }
+                    });
+                    return result;
                 }
             };
         }
